unit QSort;
{ *************************** QSORT.PAS *******************************
 PURPOSE AND IMPLEMENTATION:
  This unit provides generic quick sort function for use by various
  routines.  It also provides some comparison routines to make writing
  a sort routine easier.

 HISTORY:
   Created on June 19, 1996 by Colin Patrick Sarsfield
   Code for sort logic extracted from
    Quick Sort Routine (qsort.p) by Charlie Ross
    http://ecswww.baylor.edu/code_storehouse/pascal/qsort.html
   Updated 11/28/2001 by CPS to fix rare stack overflow bug.
   Added comment 7/3/2002 from James G. Owen -- http://home.att.net/~owen_labs/
   Revised 8/19/2002 to use QuickSort logic from
    Jon Bentley's Programming Pearls, page 121;
    Added an InsertionSort function.
    Modified Compare functions to prevent overflow.

 COPYRIGHT:
  Released into the public domain by Colin Sarsfield 1/30/97.
 ************************************************************************}

interface

type
  TCompareProc = function(const a, b: Integer): Integer of object;
  TExchangeProc = procedure(const a, b: Integer) of object;

{ main quick sort procedure }
procedure QuickSort(Low, High: Integer; const Compare: TCompareProc;
  const Exchange: TExchangeProc; CutOff: Integer = 10);

{ insertion sort procedure }
procedure InsertionSort(Low, High: Integer; const Compare: TCompareProc;
  const Exchange: TExchangeProc);

{ some useful comparison routines }
function CompareFloats(const a, b: Single): Integer;
function CompareIntegers(const a, b: Integer): Integer;

implementation

uses
  SysUtils;

function CompareFloats(const a, b: Single): Integer;
begin
  Result := Ord(a > b) - Ord(b > a);
end;

function CompareIntegers(const a, b: Integer): Integer;
begin
  Result := Ord(a > b) - Ord(b > a);
end;

procedure InsertionSort(Low, High: Integer; const Compare: TCompareProc;
  const Exchange: TExchangeProc);
var
  i, j: Integer;
begin
  for i := Succ(Low) to High do begin
    j := i;
    while (j > 0) and (Compare(j, Pred(j)) < 0) do begin
      Exchange(j, Pred(j));
      Dec(j);
    end;
  end;
end;

{ Sorts an array using an optimized QuickSort algorithm.  This algorithm favors
  sorted arrays and performs well on random arrays.  It performs poorly on fairly
  unusual arrays.  For example, these are some worst cases for CutOff = zero:
    7 3 11 5 9 1 2 4 6 8 10 12  (MaxCompares = 99, MaxExchanges = 22)
    2 10 4 8 6 1 3 5 7 9 11     (MaxCompares = 85, MaxExchanges = 20)
    9 3 7 5 1 2 4 6 8 10        (MaxCompares = 72, MaxExchanges = 18)
    2 6 4 8 1 3 5 7 9           (MaxCompares = 60, MaxExchanges = 16)

  Parameters are as follows:
    Low, High             The bounds of the array
    Compare, Exchange     The functions to compare and exchange array elements
    CutOff                An integer that specifies at what number of elements a
                            quicksort partition should be cut off, and an insertion
                            sort used to finish.  If CutOff = 0, only QuickSort
                            is used.  CutOff cannot be less than zero.  See the
                            table below for the effect of CutOff on some test
                            datasets.

  Number of Compares / Number of Exchanges using a 159 element array
  CutOff  A         B          C          D
  0       1088/655  1261/403   1260/286   1216/198
  5       958/431   1438/332   992/130    958/62
  10      798/335   1399/445   832/85     798/30
  25      638/247   1478/645   747/136    638/14

  A:  Identical values
  B:  Random values
  C:  Sorted with some identical values
  D:  Sorted with no identical values
}
procedure QuickSort(Low, High: Integer; const Compare: TCompareProc;
  const Exchange: TExchangeProc; CutOff: Integer);

  procedure QuickSub(l, h: Integer);
  var
    i, j: Integer;
  begin
    while h - l > CutOff do begin
      Exchange((l + h) shr 1, l);
      i := l; j:= Succ(h);
      while True do begin
        repeat
          Inc(i);
        until (i > h) or (Compare(i, l) >= 0);
        repeat
          Dec(j);
        until Compare(j, l) <= 0;
        if i > j then Break;
        Exchange(i, j);
      end;
      Exchange(l, j);
      QuickSub(l, Pred(j));
      l := Succ(j);
    end;
  end;
begin
  if CutOff < 0 then
    raise Exception.Create('QuickSort cutoff may not be less than zero');

  QuickSub(Low, High);
  InsertionSort(Low, High, Compare, Exchange);
end;

end.

