{:
PURPOSE AND IMPLEMENTATION:
	This unit defines some simple routines that can be used to read and write your
  own XML documents.  Be aware that these are not intended to be able to read
  data from other sources.

HISTORY:
	08/20/02 - Created by Colin Patrick Sarsfield & Andrew Cray Vogan from source
             in previous XMLReader and XMLWriter units.

DISCLAIMER:
	Use at your own risk.

COPYRIGHT:
	Released into the public domain by Colin Sarsfield 9/13/00
}
unit XMLFormat;

interface

uses
  XMLNode;

type
  EXmlFormatError = class(EXmlError);

function XmlChildToString(Node: TXmlNode; const NodeName: String;
    const NodeType: TXMLNodeType = xntElement; const Index: Integer = 0): String;
function XmlChildToBool(Node: TXmlNode; const NodeName: String;
    const NodeType: TXMLNodeType = xntElement; const Index: Integer = 0): Boolean;
function XmlChildToFloat(Node: TXmlNode; const NodeName: String;
    const NodeType: TXMLNodeType = xntElement; const Index: Integer = 0): Single;
function XmlChildToInt(Node: TXmlNode; const NodeName: String;
    const NodeType: TXMLNodeType = xntElement; const Index: Integer = 0): Integer;
function XmlChildToDate(Node: TXmlNode; const NodeName: String;
    const NodeType: TXMLNodeType = xntElement; const Index: Integer = 0): TDateTime;
function XmlIntToStr(const Value: Integer): String; overload;
function XmlIntToStr(const Value: Int64): String; overload;
function XmlFloatToStr(const Value: Extended): String; overload;
function XmlFloatToStr(const Value: Single): String; overload;
function XmlFloatToStr(const Value: Double): String; overload;
function XmlBoolToStr(const Value: Boolean): String;
function XmlCurrToStr(const Value: Currency): String;
function XmlDateToStr(const Value: TDateTime): String;
function XmlStrToDate(const Data: String): TDateTime;
function XmlStrToInt(const Data: String): Integer;
function XmlStrToInt64(const Data: String): Int64;
function XmlStrToFloat(const Data: String): Extended;
function XmlStrToCurr(const Data: String): Currency;
function XmlStrToBool(const Data: String): Boolean;

implementation

uses
  SysUtils, StrLib;

procedure RaiseXmlStringError(const Data, TypeName: String);
begin
  raise EXmlFormatError.CreateFmt('"%s" is not a valid XML %s',
      [Data, TypeName]);
end;

function XmlChildToString(Node: TXmlNode; const NodeName: String;
    const NodeType: TXMLNodeType; const Index: Integer): String;
const
  NL = #13#10;
var
  Child: TXmlNode;
  p: TXmlNode;
begin
  Child := Node.FindChild(NodeName, NodeType, Index);
  if Child = nil then
    Result := ''
  else if Child.HasChildNodes then begin
    Result := '';
    p := Child.FirstChild;
    while p <> nil do begin
      Result := Result + NL + p.Text;
      p := p.NextSibling;
    end;
    Delete(Result, 1, Length(NL));
  end else
    Result := Child.Text
end;

function XmlChildToBool(Node: TXmlNode; const NodeName: String;
    const NodeType: TXMLNodeType = xntElement; const Index: Integer = 0): Boolean;
var
  Child: TXmlNode;
begin
  Child := Node.FindChild(NodeName, NodeType, Index);
  if Child = nil then
    Result := False
  else if Child.Text = '' then
    Result := True
  else
    Result := XmlStrToBool(Child.Text)
end;

function XmlChildToFloat(Node: TXmlNode; const NodeName: String;
    const NodeType: TXMLNodeType = xntElement; const Index: Integer = 0): Single;
var
  Child: TXmlNode;
begin
  Child := Node.FindChild(NodeName, NodeType, Index);
  if Child = nil then
    Result := 0
  else
    Result := XmlStrToFloat(Child.Text)
end;

function XmlChildToInt(Node: TXmlNode; const NodeName: String;
    const NodeType: TXMLNodeType = xntElement; const Index: Integer = 0): Integer;
var
  Child: TXmlNode;
begin
  Child := Node.FindChild(NodeName, NodeType, Index);
  if Child = nil then
    Result := 0
  else
    Result := XmlStrToInt(Child.Text)
end;

function XmlChildToDate(Node: TXmlNode; const NodeName: String;
    const NodeType: TXMLNodeType = xntElement; const Index: Integer = 0): TDateTime;
var
  Child: TXmlNode;
begin
  Child := Node.FindChild(NodeName, NodeType, Index);
  if Child = nil then
    Result := 0
  else
    Result := XmlStrToDate(Child.Text)
end;

function XmlIntToStr(const Value: Integer): String;
begin
  Result := IntToStr(Value);
end;

function XmlIntToStr(const Value: Int64): String;
begin
  Result := IntToStr(Value);
end;

function XmlFloatToStr(const Value: Extended): String;
begin
  Result := Replace(FloatToStrF(Value, ffGeneral, 18, 0), DecimalSeparator, '.');
end;

function XmlFloatToStr(const Value: Single): String;
begin
  Result := Replace(FloatToStrF(Value, ffGeneral, 7, 0), DecimalSeparator, '.');
end;

function XmlFloatToStr(const Value: Double): String;
begin
  Result := Replace(FloatToStrF(Value, ffGeneral, 15, 0), DecimalSeparator, '.')
end;

function XmlBoolToStr(const Value: Boolean): String;
begin
  if Value then
    Result := 'Yes'
  else
    Result := 'No';
end;

function XmlCurrToStr(const Value: Currency): String;
type
  XForm = record
    case X: Integer of
      0: (I: Int64);
      1: (C: Currency);
    end;
var
  X: XForm;
begin
  X.C := Value;
  Result := IntToStr(X.I);
end;

function XmlDateToStr(const Value: TDateTime): String;
begin
  Result := XmlFloatToStr(Value);
end;

function XmlStrToDate(const Data: String): TDateTime;
var
  Dbl: Double;
  Code: Integer;
begin
  Val(Data, Dbl, Code);
  if Code <> 0 then
    Result := StrToDateTime(Data)
  else
    Result := TDateTime(Dbl)
end;

function XmlStrToInt(const Data: String): Integer;
begin
  try
    Result := StrToInt(Data);
  except
    RaiseXmlStringError(Data, 'Integer');
    Result := 0; // to make compiler happy
  end;
end;

function XmlStrToInt64(const Data: String): Int64;
begin
  try
    Result := StrToInt64(Data);
  except
    RaiseXmlStringError(Data, 'Int64');
    Result := 0; // to make compiler happy
  end;
end;

function XmlStrToFloat(const Data: String): Extended;
var
  Code: Integer;
begin
  Val(Data, Result, Code);
  if Code <> 0 then
    RaiseXmlStringError(Data, 'floating point value');
end;

function XmlStrToCurr(const Data: String): Currency;
type
  XForm = record
    case X: Integer of
      0: (I: Int64);
      1: (C: Currency);
    end;
var
  X: XForm;
begin
  try
    X.I := StrToInt64(Data);
    Result := X.C;
  except
    RaiseXmlStringError(Data, 'Currency');
    Result := 0; // to make compiler happy
  end;
end;

function XmlStrToBool(const Data: String): Boolean;
begin
  Result := (Length(Data) > 0) and (Data[1] in ['y', 'Y', 't', 'T'])
end;

end.
